import sys

from .utils import Command


@Command
def reload_mods(name: str):
    """Reload any python module in the current xonsh session.
    Helpful during development.

    Parameters
    ----------
    name:
        Name of the module/package to reload.
        Giving partial names matches all the nested modules.

    Examples
    -------
    $ reload-mods xontrib
        - this will reload all modules imported that starts with xontrib name

    Notes
    -----
        Please use
            `import module` or `import module as mdl` patterns
        Using
            `from module import name`
            will not reload the name imported
    """
    # todo: implement a watcher mode
    import importlib

    modules = list(
        sys.modules
    )  # create a copy of names. so that it will not raise warning
    for mod_name in modules:
        mod = sys.modules[mod_name]
        if name and not mod_name.startswith(name):
            continue
        print("reload ", mod_name)
        try:
            importlib.reload(mod)
        except Exception as e:
            print("failed to reload", mod_name, e)



import operator
import os
import typing as tp

from arger import Argument

from .utils import Command, xsh
import xonsh.tools as xt
from pathlib import Path

ENVS = {}


@xt.lazyobject
def added_file_path() -> Path:
    return (
        Path(xsh.env.get("XDG_DATA_HOME", "~/.local/share")).resolve()
        / "dev-paths.json"
    )


def update_saved_paths(paths=()):
    import json

    file = tp.cast(Path, added_file_path)
    if paths:
        content = json.dumps(list(set(paths)))
        file.write_text(content)


def get_added_paths(add_path: str = None) -> tp.Dict[str, str]:
    import json

    file = tp.cast(Path, added_file_path)

    if file.exists():
        paths = json.loads(file.read_text())
    else:
        paths = []
    if add_path:
        paths.append(add_path)
        update_saved_paths(paths)

    return {os.path.split(p)[-1]: p for p in paths}


def register_project(fn: tp.Callable = None):
    """Register new project.

    Parameters
    ----------
    call
        This function will get invoked upon finding the project_path.
        the function name will be used to search in $PROJECT_PATHS
    """

    def _wrapper():
        path = _start_proj_shell(fn.__name__)
        result = fn(path)

        return result

    ENVS[fn.__name__] = _wrapper
    return _wrapper


def _start_proj_shell(env: tp.Union[str, Path]):
    if isinstance(env, Path) and env.exists():
        path = str(env)
    else:
        path = find_proj_path(env)
        if not path:
            raise Exception("Project not found")
    os.chdir(path)
    return path


def _find_proj_path(name, op):
    for direc in xsh.env.get("PROJECT_PATH", []):
        for path in Path(direc).glob("*"):
            if op(path.name, name):
                return path


def find_proj_path(name):
    for op in [operator.eq, str.startswith, operator.contains]:
        path = _find_proj_path(name, op)
        if path:
            return path


def _list_cmds():
    from rich.console import Console

    c = Console()
    paths = get_added_paths()
    c.print("Paths", paths)


def _add_current_path():
    from rich.console import Console

    c = Console()
    path = Path.cwd().resolve()
    c.log("Adding cwd to project-paths", {path.name: path})
    get_added_paths(str(path))


@Command
def _dev(
    _arger_,
    _namespace_,
    name: tp.cast(str, Argument(nargs="?")),
    add=False,
    ls=False,
):
    """A command to cd into a directory

    Inspired from my own workflow and these commands
        - https://github.com/ohmyzsh/ohmyzsh/tree/master/plugins/pj
        - https://github.com/ohmyzsh/ohmyzsh/tree/master/plugins/wd

    Parameters
    ----------
    name
        name of the folder to cd into.
        This searches for names under $PROJECT_PATHS
        or the ones registered with `dev --add`
    add
        register the current folder to dev command.
        When using this, it will get saved in a file,
        also that is used during completions.
    ls
        show currently registered paths

    Notes
    -----
        One can use `register_project` function to regiter custom_callbacks to run on cd to project path

    Examples
    -------
        > $PROJECT_PATHS = ["~/src/"]
        > dev proj-name # --> will cd into ~/src/proj-name
        > dev --add
    """

    if ls:
        _list_cmds()
    elif add:
        _add_current_path()
    else:
        added_paths = get_added_paths()
        if name in ENVS:
            ENVS[name]()
        elif name in added_paths:

            path = added_paths[name]
            if os.path.exists(path):
                return _start_proj_shell(Path())
            else:
                # old/expired link
                added_paths.pop(name)
                update_saved_paths(tuple(added_paths.values()))

        return _start_proj_shell(name)



import builtins
import functools
import typing as tp

from xonsh.built_ins import XonshSession

xsh = tp.cast(XonshSession, builtins.__xonsh__)


class Command:
    def __init__(self, func: tp.Callable, **kwargs):
        """Convert the given function to alias and also create a argparser for its parameters"""
        dashed_name = func.__name__.strip("_").replace("_", "-")
        kwargs["func"] = func
        self.kwargs = kwargs
        self.subs = []
        # convert to
        builtins.aliases[dashed_name] = self.handle_cmd

    @property
    @functools.lru_cache()
    def parser(self):
        import arger
        import argparse

        parser = arger.Arger(
            **self.kwargs, formatter_class=argparse.RawTextHelpFormatter
        )
        for sub in self.subs:
            parser.add_cmd(sub)
        return parser

    def handle_cmd(self, args, stdout):
        self.parser.set_defaults(_stdout=stdout)
        self.parser.run(*args)

    def add(self, func):
        self.subs.append(func)
        return self


def run(*args):
    """helper function to run shell commands inside xonsh session"""
    import shlex

    cmd_args = list(args)
    if len(args) == 1 and isinstance(args[0], str) and " " in args[0]:
        first_arg = args[0]

        if " | " in first_arg:
            cmds = first_arg.split(" | ")
            cmds = map(lambda x: shlex.split(x), cmds)
            cmd_args = list(cmds)
        else:
            cmd_args = shlex.split(first_arg)
    return xsh.subproc_captured_stdout(cmd_args)
